import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids
from autolens.data.array.util import grid_util
from autolens.data.array import scaled_array as sca
from autolens.model.galaxy import galaxy as g, galaxy_data as gd
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

from test.unit.mock.model import mock_galaxy

class TestGalaxyFitData(object):

    def test__image_noise_map_and_mask(self, gal_data_5x5, mask_5x5):
        
        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, use_intensities=True)

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((5, 5))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((5, 5))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (galaxy_fit_data.mask_2d == np.array([[True, True, True, True, True],
                                                     [True, False, False, False, True],
                                                     [True, False, False, False, True],
                                                     [True, False, False, False, True],
                                                     [True,  True,  True,  True, True]])).all()

        assert (galaxy_fit_data.image_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0]])).all()


        assert (galaxy_fit_data.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                          [0.0, 2.0, 2.0, 2.0, 0.0],
                                                          [0.0, 2.0, 2.0, 2.0, 0.0],
                                                          [0.0, 2.0, 2.0, 2.0, 0.0],
                                                          [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

    def test__grid_stack(self, gal_data_5x5, mask_5x5, regular_grid_5x5, sub_grid_5x5, blurring_grid_5x5):

        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, use_intensities=True)

        assert galaxy_fit_data.grid_stack.regular.shape == (9, 2)

        assert (galaxy_fit_data.grid_stack.regular == regular_grid_5x5).all()
        assert (galaxy_fit_data.grid_stack.sub == sub_grid_5x5).all()

    def test__padded_grid_stack(self, gal_data_5x5, mask_5x5):

        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, use_intensities=True)

        padded_image_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
            mask=np.full((7, 7), False),  pixel_scales=galaxy_fit_data.unmasked_image.pixel_scales)

        assert (galaxy_fit_data.padded_grid_stack.regular == padded_image_util).all()
        assert galaxy_fit_data.padded_grid_stack.regular.image_shape == (5, 5)
        assert galaxy_fit_data.padded_grid_stack.regular.padded_shape == (7, 7)

        padded_sub_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
             mask=np.full((7,7), False), pixel_scales=galaxy_fit_data.unmasked_image.pixel_scales,
            sub_grid_size=galaxy_fit_data.grid_stack.sub.sub_grid_size)

        assert galaxy_fit_data.padded_grid_stack.sub == pytest.approx(padded_sub_util, 1e-4)
        assert galaxy_fit_data.padded_grid_stack.sub.image_shape == (5, 5)
        assert galaxy_fit_data.padded_grid_stack.sub.padded_shape == (7, 7)

    def test__interp_pixel_scale(self, image_5x5, mask_5x5):

        noise_map = sca.ScaledSquarePixelArray(array=2.0 * np.ones((5, 5)), pixel_scale=3.0)
        gal_data_5x5 = gd.GalaxyData(image=image_5x5, noise_map=noise_map, pixel_scale=3.0)
        gal_data_5x5 = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, interp_pixel_scale=1.0, use_intensities=True)

        grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
                        mask=mask_5x5, sub_grid_size=2, psf_shape=(3, 3))
        new_grid_stack = grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(interp_pixel_scale=1.0)

        assert (gal_data_5x5.grid_stack.regular == new_grid_stack.regular).all()
        assert (gal_data_5x5.grid_stack.regular.interpolator.vtx == new_grid_stack.regular.interpolator.vtx).all()
        assert (gal_data_5x5.grid_stack.regular.interpolator.wts == new_grid_stack.regular.interpolator.wts).all()

        assert (gal_data_5x5.grid_stack.sub == new_grid_stack.sub).all()
        assert (gal_data_5x5.grid_stack.sub.interpolator.vtx == new_grid_stack.sub.interpolator.vtx).all()
        assert (gal_data_5x5.grid_stack.sub.interpolator.wts == new_grid_stack.sub.interpolator.wts).all()

        assert (gal_data_5x5.grid_stack.blurring == new_grid_stack.blurring).all()
        assert (gal_data_5x5.grid_stack.blurring.interpolator.vtx == new_grid_stack.blurring.interpolator.vtx).all()
        assert (gal_data_5x5.grid_stack.blurring.interpolator.wts == new_grid_stack.blurring.interpolator.wts).all()

        padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(
            mask=mask_5x5, sub_grid_size=2, psf_shape=(3, 3))
        new_padded_grid_stack = \
            padded_grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(interp_pixel_scale=1.0)

        assert (gal_data_5x5.padded_grid_stack.regular == new_padded_grid_stack.regular).all()
        assert (
                    gal_data_5x5.padded_grid_stack.regular.interpolator.vtx == new_padded_grid_stack.regular.interpolator.vtx).all()
        assert (
                    gal_data_5x5.padded_grid_stack.regular.interpolator.wts == new_padded_grid_stack.regular.interpolator.wts).all()

        assert (gal_data_5x5.padded_grid_stack.sub == new_padded_grid_stack.sub).all()
        assert (gal_data_5x5.padded_grid_stack.sub.interpolator.vtx == new_padded_grid_stack.sub.interpolator.vtx).all()
        assert (gal_data_5x5.padded_grid_stack.sub.interpolator.wts == new_padded_grid_stack.sub.interpolator.wts).all()

    def test__gal_data_5x5_intensities(self, gal_data_5x5, mask_5x5):

        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, use_intensities=True)

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((5, 5))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((5, 5))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0* np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (galaxy_fit_data.mask_2d == np.array([[True, True,  True,  True, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True,  True,  True,  True, True]])).all()

        assert (galaxy_fit_data.image_2d ==  np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()


        assert (galaxy_fit_data.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        intensities = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (intensities == np.ones(9)).all()

        galaxy = g.Galaxy(redshift=0.5, light=lp.SphericalSersic(intensity=1.0))

        intensities_gal = galaxy.intensities_from_grid(grid=galaxy_fit_data.grid_stack.sub)
        intensities_gal = galaxy_fit_data.grid_stack.sub.regular_array_1d_from_binned_up_sub_array_1d(sub_array_1d=intensities_gal)

        intensities_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                   sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (intensities_gal == intensities_gd).all()

    def test__gal_data_5x5_convergence(self, gal_data_5x5, mask_5x5):

        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                                           use_convergence=True)

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((5, 5))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((5, 5))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0* np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (galaxy_fit_data.mask_2d == np.array([[True, True,  True,  True, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True,  True,  True,  True, True]])).all()

        assert (galaxy_fit_data.image_2d ==  np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()


        assert (galaxy_fit_data.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        convergence = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                    sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (convergence == np.ones(9)).all()

        galaxy = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0))

        convergence_gal = galaxy.convergence_from_grid(grid=galaxy_fit_data.grid_stack.sub)
        convergence_gal = galaxy_fit_data.grid_stack.sub.regular_array_1d_from_binned_up_sub_array_1d(sub_array_1d=convergence_gal)

        convergence_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                       sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (convergence_gal == convergence_gd).all()
        
    def test__gal_data_5x5_potential(self, gal_data_5x5, mask_5x5):

        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                                           use_potential=True)

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((5, 5))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((5, 5))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0* np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (galaxy_fit_data.mask_2d == np.array([[True, True,  True,  True, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True,  True,  True,  True, True]])).all()

        assert (galaxy_fit_data.image_2d ==  np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()


        assert (galaxy_fit_data.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()
        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        potential = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                              sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (potential == np.ones(9)).all()

        galaxy = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0))

        potential_gal = galaxy.potential_from_grid(grid=galaxy_fit_data.grid_stack.sub)
        potential_gal = galaxy_fit_data.grid_stack.sub.regular_array_1d_from_binned_up_sub_array_1d(sub_array_1d=potential_gal)

        potential_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                 sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (potential_gal == potential_gd).all()
        
    def test__gal_data_5x5_deflections_y(self, gal_data_5x5, mask_5x5):

        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                                           use_deflections_y=True)
        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((5, 5))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((5, 5))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0* np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (galaxy_fit_data.mask_2d == np.array([[True, True,  True,  True, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True,  True,  True,  True, True]])).all()

        assert (galaxy_fit_data.image_2d ==  np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 1.0, 1.0, 1.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()


        assert (galaxy_fit_data.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 2.0, 2.0, 2.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        deflections_y = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                  sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (deflections_y == np.ones(9)).all()

        galaxy = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0))

        deflections_gal = galaxy.deflections_from_grid(grid=galaxy_fit_data.grid_stack.sub)
        deflections_gal = np.asarray([galaxy_fit_data.grid_stack.sub.regular_array_1d_from_binned_up_sub_array_1d(deflections_gal[:, 0]),
                                      galaxy_fit_data.grid_stack.sub.regular_array_1d_from_binned_up_sub_array_1d(deflections_gal[:, 1])]).T

        deflections_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                   sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (deflections_gal[:,0] == deflections_gd).all()

    def test__gal_data_5x5_deflections_x(self, gal_data_5x5, mask_5x5):

        galaxy_fit_data = gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2, use_deflections_x=True)

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((5, 5))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((5, 5))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (galaxy_fit_data.mask_2d == np.array([[True, True, True, True, True],
                                                     [True, False, False, False, True],
                                                     [True, False, False, False, True],
                                                     [True, False, False, False, True],
                                                     [True, True, True, True, True]])).all()

        assert (galaxy_fit_data.image_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        assert (galaxy_fit_data.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                          [0.0, 2.0, 2.0, 2.0, 0.0],
                                                          [0.0, 2.0, 2.0, 2.0, 0.0],
                                                          [0.0, 2.0, 2.0, 2.0, 0.0],
                                                          [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        deflections_x = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                  sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (deflections_x == np.ones(9)).all()

        galaxy = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0))

        deflections_gal = galaxy.deflections_from_grid(grid=galaxy_fit_data.grid_stack.sub)
        deflections_gal = np.asarray([galaxy_fit_data.grid_stack.sub.regular_array_1d_from_binned_up_sub_array_1d(deflections_gal[:, 0]),
                                      galaxy_fit_data.grid_stack.sub.regular_array_1d_from_binned_up_sub_array_1d(deflections_gal[:, 1])]).T

        deflections_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(galaxies=[galaxy],
                                                                                   sub_grid=galaxy_fit_data.grid_stack.sub)

        assert (deflections_gal[:,1] == deflections_gd).all()

    def test__no_use_method__raises_exception(self, image_5x5, mask_5x5):

        gal_data_5x5 = gd.GalaxyData(image=image_5x5, noise_map=2.0*np.ones((5, 5)), pixel_scale=3.0)

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2)

    def test__multiple_use_methods__raises_exception(self, image_5x5, mask_5x5):

        gal_data_5x5 = gd.GalaxyData(image=image_5x5, noise_map=2.0*np.ones((5, 5)), pixel_scale=3.0)

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                             use_intensities=True, use_convergence=True)

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                             use_intensities=True, use_potential=True)

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                             use_intensities=True, use_deflections_y=True)

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                             use_intensities=True, use_convergence=True, use_potential=True)

        with pytest.raises(exc.GalaxyException):
                gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2,
                                 use_intensities=True, use_convergence=True, use_potential=True, use_deflections_x=True)


import numpy as np
import pytest

from autolens import exc, dimensions as dim
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

from test.unit.mock.model import mock_cosmology

class TestLightProfiles(object):

    class TestIntensity:

        def test__galaxies_with_x1_and_x2_light_profiles__intensities_is_sum_of_profiles(self, lp_0, lp_1,
                                                                                         gal_x1_lp, gal_x2_lp):

            lp_intensity = lp_0.intensities_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_lp_intensity = gal_x1_lp.intensities_from_grid(np.array([[1.05, -0.55]]))

            assert lp_intensity == gal_lp_intensity

            intensity = lp_0.intensities_from_grid(np.array([[1.05, -0.55]]))
            intensity += lp_1.intensities_from_grid(np.array([[1.05, -0.55]]))

            gal_intensity = gal_x2_lp.intensities_from_grid(np.array([[1.05, -0.55]]))

            assert intensity == gal_intensity

    class TestLuminosityWithin:

        def test__two_profile_galaxy__is_sum_of_individual_profiles(self, lp_0, lp_1, gal_x1_lp, gal_x2_lp):

            radius = dim.Length(0.5, 'arcsec')

            lp_luminosity = lp_0.luminosity_within_circle_in_units(radius=radius, unit_luminosity='eps')
            gal_luminosity = gal_x1_lp.luminosity_within_circle_in_units(radius=radius, unit_luminosity='eps')

            assert lp_luminosity == gal_luminosity

            lp_luminosity = lp_0.luminosity_within_ellipse_in_units(major_axis=radius, unit_luminosity='eps')
            lp_luminosity += lp_1.luminosity_within_ellipse_in_units(major_axis=radius, unit_luminosity='eps')

            gal_luminosity = gal_x2_lp.luminosity_within_ellipse_in_units(major_axis=radius, unit_luminosity='eps')

            assert lp_luminosity == gal_luminosity

        def test__radius_unit_conversions__multiply_by_kpc_per_arcsec(self, lp_0, gal_x1_lp):

            cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0)

            radius = dim.Length(0.5, 'arcsec')

            lp_luminosity_arcsec = lp_0.luminosity_within_circle_in_units(radius=radius)
            gal_luminosity_arcsec = gal_x1_lp.luminosity_within_circle_in_units(radius=radius)

            assert lp_luminosity_arcsec == gal_luminosity_arcsec

            radius = dim.Length(0.5, 'kpc')

            lp_luminosity_kpc = lp_0.luminosity_within_circle_in_units(
                radius=radius, redshift_profile=0.5,cosmology=cosmology)
            gal_luminosity_kpc = gal_x1_lp.luminosity_within_circle_in_units(radius=radius, cosmology=cosmology)

            assert lp_luminosity_kpc == gal_luminosity_kpc

        def test__luminosity_unit_conversions__multiply_by_exposure_time(self, lp_0, gal_x1_lp):

            radius = dim.Length(0.5, 'arcsec')

            lp_luminosity_eps = lp_0.luminosity_within_ellipse_in_units(
                major_axis=radius, unit_luminosity='eps', exposure_time=2.0)
            gal_luminosity_eps = gal_x1_lp.luminosity_within_ellipse_in_units(
                major_axis=radius, unit_luminosity='eps',exposure_time=2.0)

            assert lp_luminosity_eps == gal_luminosity_eps

            lp_luminosity_counts = lp_0.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity='counts', exposure_time=2.0)

            gal_luminosity_counts = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity='counts', exposure_time=2.0)

            assert lp_luminosity_counts == gal_luminosity_counts

        def test__no_light_profile__returns_none(self):

            gal_no_lp = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal())

            assert gal_no_lp.luminosity_within_circle_in_units(radius=1.0) == None
            assert gal_no_lp.luminosity_within_ellipse_in_units(major_axis=1.0) == None

    class TestSymmetricProfiles(object):

        def test_1d_symmetry(self):

            lp_0 = lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6, sersic_index=4.0)

            lp_1 = lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6, sersic_index=4.0,
                                       centre=(100, 0))

            gal_x2_lp = g.Galaxy(redshift=0.5, light_profile_0=lp_0, light_profile_1=lp_1)

            assert gal_x2_lp.intensities_from_grid(
                np.array([[0.0, 0.0]])) == gal_x2_lp.intensities_from_grid(np.array([[100.0, 0.0]]))
            assert gal_x2_lp.intensities_from_grid(
                np.array([[49.0, 0.0]])) == gal_x2_lp.intensities_from_grid(np.array([[51.0, 0.0]]))

        def test_2d_symmetry(self):

            lp_0 = lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0,effective_radius=0.6, sersic_index=4.0)

            lp_1 = lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6, sersic_index=4.0,
                                       centre=(100, 0))

            lp_2 = lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6, sersic_index=4.0,
                                       centre=(0, 100))

            lp_3 = lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6, sersic_index=4.0,
                                       centre=(100, 100))

            gal_x4_lp = g.Galaxy(redshift=0.5, light_profile_0=lp_0, light_profile_1=lp_1,
                                  light_profile_3=lp_2, light_profile_4=lp_3)

            assert gal_x4_lp.intensities_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                gal_x4_lp.intensities_from_grid(np.array([[51.0, 0.0]])), 1e-5)

            assert gal_x4_lp.intensities_from_grid(np.array([[0.0, 49.0]])) == pytest.approx(
                gal_x4_lp.intensities_from_grid(np.array([[0.0, 51.0]])), 1e-5)

            assert gal_x4_lp.intensities_from_grid(np.array([[100.0, 49.0]])) == pytest.approx(
                gal_x4_lp.intensities_from_grid(np.array([[100.0, 51.0]])), 1e-5)

            assert gal_x4_lp.intensities_from_grid(np.array([[49.0, 49.0]])) == pytest.approx(
                gal_x4_lp.intensities_from_grid(np.array([[51.0, 51.0]])), 1e-5)


class TestMassProfiles(object):

    class TestConvergence:

        def test__galaxies_with_x1_and_x2_mass_profiles__convergence_is_same_individual_profiles(self, mp_0, gal_x1_mp,
                                                                                                 mp_1, gal_x2_mp):

            mp_convergence = mp_0.convergence_from_grid(np.array([[1.05, -0.55]]))

            gal_mp_convergence = gal_x1_mp.convergence_from_grid(np.array([[1.05, -0.55]]))

            assert mp_convergence == gal_mp_convergence

            mp_convergence = mp_0.convergence_from_grid(np.array([[1.05, -0.55]]))
            mp_convergence += mp_1.convergence_from_grid(np.array([[1.05, -0.55]]))

            gal_convergence = gal_x2_mp.convergence_from_grid(np.array([[1.05, -0.55]]))

            assert mp_convergence == gal_convergence

    class TestPotential:

        def test__galaxies_with_x1_and_x2_mass_profiles__potential_is_same_individual_profiles(self, mp_0, gal_x1_mp,
                                                                                               mp_1, gal_x2_mp):

            mp_potential = mp_0.potential_from_grid(np.array([[1.05, -0.55]]))

            gal_mp_potential = gal_x1_mp.potential_from_grid(np.array([[1.05, -0.55]]))

            assert mp_potential == gal_mp_potential

            mp_potential = mp_0.potential_from_grid(np.array([[1.05, -0.55]]))
            mp_potential += mp_1.potential_from_grid(np.array([[1.05, -0.55]]))

            gal_potential = gal_x2_mp.potential_from_grid(np.array([[1.05, -0.55]]))

            assert mp_potential == gal_potential

    class TestDeflectionAngles:

        def test__galaxies_with_x1_and_x2_mass_profiles__deflection_same_as_individual_profiles(self, mp_0, gal_x1_mp,
                                                                                                mp_1, gal_x2_mp):

            mp_deflections = mp_0.deflections_from_grid(np.array([[1.05, -0.55]]))

            gal_mp_deflections = gal_x1_mp.deflections_from_grid(np.array([[1.05, -0.55]]))

            assert mp_deflections[0, 0] == gal_mp_deflections[0, 0]
            assert mp_deflections[0, 1] == gal_mp_deflections[0, 1]

            mp_deflections_0 = mp_0.deflections_from_grid(np.array([[1.05, -0.55]]))
            mp_deflections_1 = mp_1.deflections_from_grid(np.array([[1.05, -0.55]]))

            mp_deflections = mp_deflections_0 + mp_deflections_1

            gal_mp_deflections = gal_x2_mp.deflections_from_grid(np.array([[1.05, -0.55]]))

            assert mp_deflections[0, 0] == gal_mp_deflections[0, 0]
            assert mp_deflections[0, 1] == gal_mp_deflections[0, 1]

    class TestMassWithin:

        def test__two_profile_galaxy__is_sum_of_individual_profiles(self, mp_0, gal_x1_mp, mp_1, gal_x2_mp):

            radius = dim.Length(0.5, 'arcsec')

            mp_mass = mp_0.mass_within_circle_in_units(radius=radius, unit_mass='angular')

            gal_mass = gal_x1_mp.mass_within_circle_in_units(radius=radius, unit_mass='angular')

            assert mp_mass == gal_mass

            mp_mass = mp_0.mass_within_ellipse_in_units(major_axis=radius, unit_mass='angular')
            mp_mass += mp_1.mass_within_ellipse_in_units(major_axis=radius, unit_mass='angular')

            gal_mass = gal_x2_mp.mass_within_ellipse_in_units(major_axis=radius, unit_mass='angular')

            assert mp_mass == gal_mass

        def test__radius_unit_conversions__multiply_by_kpc_per_arcsec(self, mp_0, gal_x1_mp):

            cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=1.0)

            radius = dim.Length(0.5, 'arcsec')

            mp_mass_arcsec = mp_0.mass_within_circle_in_units(radius=radius, unit_mass='solMass', redshift_profile=0.5,
                                                   redshift_source=1.0, cosmology=cosmology)

            gal_mass_arcsec = gal_x1_mp.mass_within_circle_in_units(radius=radius, unit_mass='solMass',
                                                                      redshift_source=1.0, cosmology=cosmology)
            assert mp_mass_arcsec == gal_mass_arcsec

            radius = dim.Length(0.5, 'kpc')

            mp_mass_kpc = mp_0.mass_within_circle_in_units(radius=radius, unit_mass='solMass', redshift_profile=0.5,
                                                   redshift_source=1.0, cosmology=cosmology)

            gal_mass_kpc = gal_x1_mp.mass_within_circle_in_units(radius=radius, unit_mass='solMass',
                                                                   redshift_source=1.0, cosmology=cosmology)
            assert mp_mass_kpc == gal_mass_kpc

        def test__mass_unit_conversions__same_as_individual_profile(self, mp_0, gal_x1_mp):

            cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=1.0, kpc_per_arcsec=1.0,
                                                     critical_surface_density=2.0)

            radius = dim.Length(0.5, 'arcsec')

            mp_mass_angular = mp_0.mass_within_ellipse_in_units(major_axis=radius, unit_mass='angular',
                                                    redshift_profile=0.5, redshift_source=1.0, cosmology=cosmology)

            gal_mass_angular = gal_x1_mp.mass_within_ellipse_in_units(major_axis=radius, unit_mass='angular',
                                                           redshift_source=1.0, cosmology=cosmology)
            assert mp_mass_angular == gal_mass_angular

            mp_mass_sol = mp_0.mass_within_circle_in_units(radius=radius, unit_mass='solMass',
                                                   redshift_profile=0.5, redshift_source=1.0, cosmology=cosmology)

            gal_mass_sol = gal_x1_mp.mass_within_circle_in_units(radius=radius, unit_mass='solMass',
                                                        redshift_source=1.0, cosmology=cosmology)
            assert mp_mass_sol == gal_mass_sol

        def test__no_mass_profile__returns_none(self):

            gal_no_mp = g.Galaxy(redshift=0.5, light=lp.SphericalSersic())

            assert gal_no_mp.mass_within_circle_in_units(radius=1.0, critical_surface_density=1.0) == None
            assert gal_no_mp.mass_within_ellipse_in_units(major_axis=1.0, critical_surface_density=1.0) == None

    class TestSymmetricProfiles:

        def test_1d_symmetry(self):

            mp_0 = mp.EllipticalIsothermal(axis_ratio=0.5, phi=45.0,
                                                   einstein_radius=1.0)

            mp_1 = mp.EllipticalIsothermal(centre=(100, 0), axis_ratio=0.5, phi=45.0,
                                                   einstein_radius=1.0)

            gal_x4_mp = g.Galaxy(redshift=0.5, mass_profile_0=mp_0, mass_profile_1=mp_1)

            assert gal_x4_mp.convergence_from_grid(
                np.array([[1.0, 0.0]])) == gal_x4_mp.convergence_from_grid(np.array([[99.0, 0.0]]))

            assert gal_x4_mp.convergence_from_grid(
                np.array([[49.0, 0.0]])) == gal_x4_mp.convergence_from_grid(np.array([[51.0, 0.0]]))

            assert gal_x4_mp.potential_from_grid(np.array([[1.0, 0.0]])) == pytest.approx(
                gal_x4_mp.potential_from_grid(np.array([[99.0, 0.0]])), 1e-6)

            assert gal_x4_mp.potential_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                gal_x4_mp.potential_from_grid(np.array([[51.0, 0.0]])), 1e-6)

            assert gal_x4_mp.deflections_from_grid(np.array([[1.0, 0.0]])) == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[99.0, 0.0]])), 1e-6)

            assert gal_x4_mp.deflections_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[51.0, 0.0]])), 1e-6)

        def test_2d_symmetry(self):

            mp_0 = mp.SphericalIsothermal(einstein_radius=1.0)

            mp_1 = mp.SphericalIsothermal(centre=(100, 0), einstein_radius=1.0)

            mp_2 = mp.SphericalIsothermal(centre=(0, 100), einstein_radius=1.0)

            mp_3 = mp.SphericalIsothermal(centre=(100, 100), einstein_radius=1.0)

            gal_x4_mp = g.Galaxy(redshift=0.5,
                                      mass_profile_0=mp_0, mass_profile_1=mp_1,
                                      mass_profile_2=mp_2, mass_profile_3=mp_3)

            assert gal_x4_mp.convergence_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                gal_x4_mp.convergence_from_grid(np.array([[51.0, 0.0]])), 1e-5)

            assert gal_x4_mp.convergence_from_grid(np.array([[0.0, 49.0]])) == pytest.approx(
                gal_x4_mp.convergence_from_grid(np.array([[0.0, 51.0]])), 1e-5)

            assert gal_x4_mp.convergence_from_grid(np.array([[100.0, 49.0]])) == pytest.approx(
                gal_x4_mp.convergence_from_grid(np.array([[100.0, 51.0]])), 1e-5)

            assert gal_x4_mp.convergence_from_grid(np.array([[49.0, 49.0]])) == pytest.approx(
                gal_x4_mp.convergence_from_grid(np.array([[51.0, 51.0]])), 1e-5)

            assert gal_x4_mp.potential_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                gal_x4_mp.potential_from_grid(np.array([[51.0, 0.0]])), 1e-5)

            assert gal_x4_mp.potential_from_grid(np.array([[0.0, 49.0]])) == pytest.approx(
                gal_x4_mp.potential_from_grid(np.array([[0.0, 51.0]])), 1e-5)

            assert gal_x4_mp.potential_from_grid(np.array([[100.0, 49.0]])) == pytest.approx(
                gal_x4_mp.potential_from_grid(np.array([[100.0, 51.0]])), 1e-5)

            assert gal_x4_mp.potential_from_grid(np.array([[49.0, 49.0]])) == pytest.approx(
                gal_x4_mp.potential_from_grid(np.array([[51.0, 51.0]])), 1e-5)

            assert -1.0 * gal_x4_mp.deflections_from_grid(np.array([[49.0, 0.0]]))[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[51.0, 0.0]]))[0, 0], 1e-5)

            assert 1.0 * gal_x4_mp.deflections_from_grid(np.array([[0.0, 49.0]]))[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[0.0, 51.0]]))[0, 0], 1e-5)

            assert 1.0 * gal_x4_mp.deflections_from_grid(np.array([[100.0, 49.0]]))[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[100.0, 51.0]]))[0, 0], 1e-5)

            assert -1.0 * gal_x4_mp.deflections_from_grid(np.array([[49.0, 49.0]]))[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[51.0, 51.0]]))[0, 0], 1e-5)

            assert 1.0 * gal_x4_mp.deflections_from_grid(np.array([[49.0, 0.0]]))[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[51.0, 0.0]]))[0, 1], 1e-5)

            assert -1.0 * gal_x4_mp.deflections_from_grid(np.array([[0.0, 49.0]]))[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[0.0, 51.0]]))[0, 1], 1e-5)

            assert -1.0 * gal_x4_mp.deflections_from_grid(np.array([[100.0, 49.0]]))[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[100.0, 51.0]]))[0, 1], 1e-5)

            assert -1.0 * gal_x4_mp.deflections_from_grid(np.array([[49.0, 49.0]]))[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(np.array([[51.0, 51.0]]))[0, 1], 1e-5)

    class TestEinsteinRadiiMass:

        def test__x2_sis_different_einstein_radii_and_mass__einstein_radii_and_mass_are_sum(self):

            mp_0 = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
            mp_1 = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=0.5)

            gal_x2_mp = g.Galaxy(redshift=0.5, mass_0=mp_0, mass_1=mp_1)

            assert gal_x2_mp.einstein_radius_in_units(unit_length='arcsec') == 1.5
            assert gal_x2_mp.einstein_mass_in_units(unit_mass='angular') == np.pi*(1.0 + 0.5**2.0)

        def test__includes_shear__does_not_impact_values(self):

            mp_0 = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
            shear = mp.ExternalShear()

            gal_shear = g.Galaxy(redshift=0.5, mass_0=mp_0, shear=shear)

            assert gal_shear.einstein_radius_in_units(unit_length='arcsec') == 1.0
            assert gal_shear.einstein_mass_in_units(unit_mass='angular') == np.pi


class TestMassAndLightProfiles(object):

    def test_single_profile(self, lmp_0):

        gal_x1_lmp = g.Galaxy(redshift=0.5, profile=lmp_0)

        assert 1 == len(gal_x1_lmp.light_profiles)
        assert 1 == len(gal_x1_lmp.mass_profiles)

        assert gal_x1_lmp.mass_profiles[0] == lmp_0
        assert gal_x1_lmp.light_profiles[0] == lmp_0

    def test_multiple_profile(self, lmp_0, lp_0, mp_0):

        gal_multi_profiles = g.Galaxy(redshift=0.5, profile=lmp_0, light=lp_0, sie=mp_0)

        assert 2 == len(gal_multi_profiles.light_profiles)
        assert 2 == len(gal_multi_profiles.mass_profiles)


class TestSummarizeInUnits(object):

    def test__galaxy_with_two_light_and_mass_profiles(self, lp_0, lp_1, mp_0, mp_1):

        gal_summarize = g.Galaxy(redshift=0.5, light_profile_0=lp_0, light_profile_1=lp_1,
                       mass_profile_0=mp_0, mass_profile_1=mp_1)


        summary_text = gal_summarize.summarize_in_units(radii=[dim.Length(10.0), dim.Length(500.0)], whitespace=50,
                                              unit_length='arcsec', unit_luminosity='eps', unit_mass='angular')

        i = 0

        assert summary_text[i] == 'Galaxy\n' ; i += 1
        assert summary_text[i] == 'redshift                                          0.50' ; i += 1
        assert summary_text[i] == '\nGALAXY LIGHT\n\n' ; i += 1
        assert summary_text[i] == 'luminosity_within_10.00_arcsec                    1.8854e+02 eps' ; i += 1
        assert summary_text[i] == 'luminosity_within_500.00_arcsec                   1.9573e+02 eps' ; i += 1
        assert summary_text[i] ==  '\nLIGHT PROFILES:\n\n' ; i += 1
        assert summary_text[i] == 'Light Profile = SphericalSersic\n' ; i += 1
        assert summary_text[i] == 'luminosity_within_10.00_arcsec                    6.2848e+01 eps' ; i += 1
        assert summary_text[i] == 'luminosity_within_500.00_arcsec                   6.5243e+01 eps' ; i += 1
        assert summary_text[i] == '\n' ; i += 1
        assert summary_text[i] == 'Light Profile = SphericalSersic\n' ; i += 1
        assert summary_text[i] == 'luminosity_within_10.00_arcsec                    1.2570e+02 eps' ; i += 1
        assert summary_text[i] == 'luminosity_within_500.00_arcsec                   1.3049e+02 eps' ; i += 1
        assert summary_text[i] == '\n' ; i += 1
        assert summary_text[i] ==  '\nGALAXY MASS\n\n' ; i += 1
        assert summary_text[i] == 'einstein_radius                                   3.00 arcsec' ; i += 1
        assert summary_text[i] == 'einstein_mass                                     1.5708e+01 angular' ; i += 1
        assert summary_text[i] == 'mass_within_10.00_arcsec                          9.4248e+01 angular' ; i += 1
        assert summary_text[i] == 'mass_within_500.00_arcsec                         4.7124e+03 angular' ; i += 1
        assert summary_text[i] ==  '\nMASS PROFILES:\n\n' ; i += 1
        assert summary_text[i] == 'Mass Profile = SphericalIsothermal\n' ; i += 1
        assert summary_text[i] == 'einstein_radius                                   1.00 arcsec' ; i += 1
        assert summary_text[i] == 'einstein_mass                                     3.1416e+00 angular' ; i += 1
        assert summary_text[i] == 'mass_within_10.00_arcsec                          3.1416e+01 angular' ; i += 1
        assert summary_text[i] == 'mass_within_500.00_arcsec                         1.5708e+03 angular' ; i += 1
        assert summary_text[i] == '\n' ; i += 1
        assert summary_text[i] == 'Mass Profile = SphericalIsothermal\n' ; i += 1
        assert summary_text[i] == 'einstein_radius                                   2.00 arcsec' ; i += 1
        assert summary_text[i] == 'einstein_mass                                     1.2566e+01 angular' ; i += 1
        assert summary_text[i] == 'mass_within_10.00_arcsec                          6.2832e+01 angular' ; i += 1
        assert summary_text[i] == 'mass_within_500.00_arcsec                         3.1416e+03 angular' ; i += 1
        assert summary_text[i] == '\n' ; i += 1


class TestHyperGalaxy(object):

    class TestContributionMaps(object):

        def test__model_image_all_1s__factor_is_0__contributions_all_1s(self):

            hyper_image = np.ones((3,))

            hyp = g.HyperGalaxy(contribution_factor=0.0)
            contribution_map = hyp.contribution_map_from_hyper_images(
                hyper_model_image=hyper_image, hyper_galaxy_image=hyper_image)

            assert (contribution_map == np.ones((3,))).all()

        def test__different_values__factor_is_1__contributions_are_value_divided_by_factor_and_max(self):

            hyper_image = np.array([0.5, 1.0, 1.5])

            hyp = g.HyperGalaxy(contribution_factor=1.0)
            contribution_map = hyp.contribution_map_from_hyper_images(
                hyper_model_image=hyper_image, hyper_galaxy_image=hyper_image)

            assert (contribution_map == np.array([(0.5 / 1.5) / (1.5 / 2.5), (1.0 / 2.0) / (1.5 / 2.5), 1.0])).all()

    class TestHyperNoiseMap(object):

        def test__contribution_all_1s__noise_factor_2__noise_adds_double(self):

            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.ones((3, 1))

            hyper_galaxy = g.HyperGalaxy(contribution_factor=0.0, noise_factor=2.0, noise_power=1.0)

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map)

            assert (hyper_noise_map == np.array([2.0, 4.0, 6.0])).all()

        def test__same_as_above_but_contributions_vary(self):

            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.array([[0.0, 0.5, 1.0]])

            hyper_galaxy = g.HyperGalaxy(contribution_factor=0.0, noise_factor=2.0, noise_power=1.0)

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map)

            assert (hyper_noise_map == np.array([0.0, 2.0, 6.0])).all()

        def test__same_as_above_but_change_noise_scale_terms(self):

            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.array([[0.0, 0.5, 1.0]])

            hyper_galaxy = g.HyperGalaxy(contribution_factor=0.0, noise_factor=2.0, noise_power=2.0)

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map)

            assert (hyper_noise_map == np.array([0.0, 2.0, 18.0])).all()


class TestUsesBools(object):

    def test__uses_cluster_inversion__tests_depend_on_any_pixelizations(self):

        galaxy = g.Galaxy(
            redshift=0.5)
        assert galaxy.uses_inversion == False

        galaxy = g.Galaxy(
            redshift=0.5,
            pixelization=pix.Rectangular(),
            regularization=reg.Constant())

        assert galaxy.uses_inversion == True

        galaxy = g.Galaxy(
            redshift=0.5,
            pixelization=pix.VoronoiBrightnessImage(),
            regularization=reg.AdaptiveBrightness())

        assert galaxy.uses_inversion == True

    def test__uses_cluster_inversion__tests_depend_specific_pixelizations(self):

        galaxy = g.Galaxy(
            redshift=0.5)
        assert galaxy.uses_cluster_inversion == False

        galaxy = g.Galaxy(
            redshift=0.5,
            pixelization=pix.Rectangular(),
            regularization=reg.Constant())

        assert galaxy.uses_cluster_inversion == False

        galaxy = g.Galaxy(
            redshift=0.5,
            pixelization=pix.VoronoiBrightnessImage(),
            regularization=reg.AdaptiveBrightness())

        assert galaxy.uses_cluster_inversion == True

    def test__uses_hyper_images__tests_depend_on_hyper_galaxy_and_specific_pixelizations_and_regularizations(self):
        
        galaxy = g.Galaxy(
            redshift=0.5)

        assert galaxy.uses_hyper_images == False

        galaxy = g.Galaxy(
            redshift=0.5,
            hyper_galaxy=g.HyperGalaxy())

        assert galaxy.uses_hyper_images == True

        galaxy = g.Galaxy(
            redshift=0.5,
            pixelization=pix.Rectangular(),
            regularization=reg.Constant())

        assert galaxy.uses_hyper_images == False

        galaxy = g.Galaxy(
            redshift=0.5,
            pixelization=pix.Rectangular(),
            regularization=reg.AdaptiveBrightness())

        assert galaxy.uses_hyper_images == True


class TestBooleanProperties(object):

    def test_has_profile(self):

        assert g.Galaxy(redshift=0.5).has_profile is False
        assert g.Galaxy(redshift=0.5, light_profile=lp.LightProfile()).has_profile is True
        assert g.Galaxy(redshift=0.5, mass_profile=mp.MassProfile()).has_profile is True

    def test_has_light_profile(self):
        assert g.Galaxy(redshift=0.5).has_light_profile is False
        assert g.Galaxy(redshift=0.5, light_profile=lp.LightProfile()).has_light_profile is True
        assert g.Galaxy(redshift=0.5, mass_profile=mp.MassProfile()).has_light_profile is False

    def test_has_mass_profile(self):
        assert g.Galaxy(redshift=0.5).has_mass_profile is False
        assert g.Galaxy(redshift=0.5, light_profile=lp.LightProfile()).has_mass_profile is False
        assert g.Galaxy(redshift=0.5, mass_profile=mp.MassProfile()).has_mass_profile is True

    def test_has_redshift(self):

        assert g.Galaxy(redshift=0.1).has_redshift is True

    def test_has_pixelization(self):
        assert g.Galaxy(redshift=0.5).has_pixelization is False
        assert g.Galaxy(redshift=0.5, pixelization=object(), regularization=object()).has_pixelization is True

    def test_has_regularization(self):
        assert g.Galaxy(redshift=0.5).has_regularization is False
        assert g.Galaxy(redshift=0.5, pixelization=object(), regularization=object()).has_regularization is True

    def test_has_hyper_galaxy(self):
        assert g.Galaxy(redshift=0.5).has_pixelization is False
        assert g.Galaxy(redshift=0.5, hyper_galaxy=object()).has_hyper_galaxy is True

    def test__only_pixelization_raises_error(self):
        with pytest.raises(exc.GalaxyException):
            g.Galaxy(redshift=0.5, pixelization=object())

    def test__only_regularization_raises_error(self):
        with pytest.raises(exc.GalaxyException):
            g.Galaxy(redshift=0.5, regularization=object())




from autolens.data.array import grids
from autolens.data.convolution import ConvolverImage
from autolens.data.array import mask as msk
from autolens.model.inversion import convolution as inversion_convolution

class LensData(object):

    def __init__(self, ccd_data, mask, sub_grid_size=2, positions=None, image_psf_shape=None, inversion_psf_shape=None,
                 interp_pixel_scale=None, cluster_pixel_scale=None, cluster_pixel_limit=None, uses_inversion=True,
                 uses_cluster_inversion=True):
        """
        The lens data is the collection of data (image, noise-map, PSF), a mask, grid_stack, convolver \
        and other utilities that are used for modeling and fitting an image of a strong lens.

        Whilst the image, noise-map, etc. are loaded in 2D, the lens data creates reduced 1D arrays of each \
        for lensing calculations.

        Parameters
        ----------
        ccd_data: im.CCD
            The ccd data all in 2D (the image, noise-map, PSF, etc.)
        mask: msk.Mask
            The 2D mask that is applied to the image.
        sub_grid_size : int
            The size of the sub-grid used for each lens SubGrid. E.g. a value of 2 grid_stack each image-pixel on a 2x2 \
            sub-grid.
        image_psf_shape : (int, int)
            The shape of the PSF used for convolving model image generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input image PSF, giving a faster analysis run-time.
        inversion_psf_shape : (int, int)
            The shape of the PSF used for convolving the inversion mapping matrix. A smaller \
            shape will trim the PSF relative to the input image PSF, giving a faster analysis run-time.
        positions : [[]]
            Lists of image-pixel coordinates (arc-seconds) that mappers close to one another in the source-plane(s), \
            used to speed up the non-linear sampling.
        interp_pixel_scale : float
            If *True*, expensive to compute mass profile deflection angles will be computed on a sparse grid and \
            interpolated to the regular, sub and blurring grids.
        cluster_pixel_limit : int or None
            The maximum number of pixels that can be used by an inversion, with the limit placed primarily to speed \
            up run.
        cluster_pixel_scale : float or None
            If *True*, the hyper image used to generate the cluster'grids weight map will be binned up to this higher \
            pixel scale to speed up the KMeans clustering algorithm.
        """

        self.ccd_data = ccd_data

        self.unmasked_image = ccd_data.image
        self.unmasked_noise_map = ccd_data.noise_map
        self.pixel_scale = ccd_data.pixel_scale
        self.psf = ccd_data.psf
        self.mask_1d = mask.map_2d_array_to_masked_1d_array(array_2d=mask)
        self.image_1d = mask.map_2d_array_to_masked_1d_array(array_2d=ccd_data.image)
        self.noise_map_1d = mask.map_2d_array_to_masked_1d_array(array_2d=ccd_data.noise_map)

        self.sub_grid_size = sub_grid_size

        if image_psf_shape is None:
            self.image_psf_shape = self.psf.shape
        else:
            self.image_psf_shape = image_psf_shape

        self.convolver_image = ConvolverImage(
            mask=mask, blurring_mask=mask.blurring_mask_for_psf_shape(psf_shape=self.image_psf_shape),
            psf=self.psf.resized_scaled_array_from_array(new_shape=self.image_psf_shape))

        if inversion_psf_shape is None:
            self.inversion_psf_shape = self.psf.shape
        else:
            self.inversion_psf_shape = inversion_psf_shape

        self.uses_inversion = uses_inversion

        if uses_inversion:

            self.convolver_mapping_matrix = inversion_convolution.ConvolverMappingMatrix(
                 mask=mask, psf=self.psf.resized_scaled_array_from_array(new_shape=self.inversion_psf_shape))

        else:

            self.convolver_mapping_matrix = None

        self.grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
            mask=mask, sub_grid_size=sub_grid_size, psf_shape=self.image_psf_shape)

        self.padded_grid_stack = grids.GridStack.padded_grid_stack_from_mask_sub_grid_size_and_psf_shape(
            mask=mask, sub_grid_size=sub_grid_size, psf_shape=self.image_psf_shape)

        self.interp_pixel_scale = interp_pixel_scale

        if interp_pixel_scale is not None:

            self.grid_stack = self.grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(
                interp_pixel_scale=interp_pixel_scale)

            self.padded_grid_stack = self.padded_grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(
                interp_pixel_scale=interp_pixel_scale)

        self.border = grids.RegularGridBorder.from_mask(mask=mask)

        self.positions = positions

        self.mask_2d = mask
        self.image_2d = self.map_to_scaled_array(array_1d=self.image_1d)
        self.noise_map_2d = self.map_to_scaled_array(array_1d=self.noise_map_1d)

        self.uses_cluster_inversion = uses_cluster_inversion

        if uses_cluster_inversion:

            self.cluster_pixel_limit = cluster_pixel_limit
            self.cluster_pixel_scale = cluster_pixel_scale

            if self.cluster_pixel_scale is not None:

                self.cluster = grids.ClusterGrid.from_mask_and_cluster_pixel_scale(
                    mask=self.mask_2d, cluster_pixel_scale=cluster_pixel_scale, cluster_pixels_limit=cluster_pixel_limit)

            else:

                self.cluster = grids.ClusterGrid.from_mask_and_cluster_pixel_scale(
                    mask=self.mask_2d, cluster_pixel_scale=self.pixel_scale, cluster_pixels_limit=cluster_pixel_limit)

        else:

            self.cluster = None
            self.cluster_pixel_limit = None
            self.cluster_pixel_scale = None


    def new_lens_data_with_modified_image(self, modified_image):

        ccd_data_with_modified_image = self.ccd_data.new_ccd_data_with_modified_image(modified_image=modified_image)

        return LensData(
            ccd_data=ccd_data_with_modified_image,
            mask=self.mask_2d,
            sub_grid_size=self.sub_grid_size,
            positions=self.positions,
            image_psf_shape=self.image_psf_shape,
            inversion_psf_shape=self.inversion_psf_shape,
            interp_pixel_scale=self.interp_pixel_scale,
            cluster_pixel_scale=self.cluster_pixel_scale,
            cluster_pixel_limit=self.cluster_pixel_limit,
            uses_inversion=self.uses_inversion,
            uses_cluster_inversion=self.uses_cluster_inversion)

    def new_lens_data_with_binned_up_ccd_data_and_mask(self, bin_up_factor):

        binned_up_ccd_data = self.ccd_data.new_ccd_data_with_binned_up_arrays(bin_up_factor=bin_up_factor)
        binned_up_mask = self.mask_2d.binned_up_mask_from_mask(bin_up_factor=bin_up_factor)

        return LensData(
            ccd_data=binned_up_ccd_data,
            mask=binned_up_mask,
            sub_grid_size=self.sub_grid_size,
            positions=self.positions,
            image_psf_shape=self.image_psf_shape,
            inversion_psf_shape=self.inversion_psf_shape,
            interp_pixel_scale=self.interp_pixel_scale,
            cluster_pixel_scale=self.cluster_pixel_scale,
            cluster_pixel_limit=self.cluster_pixel_limit,
            uses_inversion=self.uses_inversion,
            uses_cluster_inversion=self.uses_cluster_inversion)

    @property
    def array_1d_from_array_2d(self):
        return self.grid_stack.regular.array_1d_from_array_2d

    @property
    def map_to_scaled_array(self):
        return self.grid_stack.scaled_array_2d_from_array_1d

    def __array_finalize__(self, obj):
        if isinstance(obj, LensData):
            self.ccd_data = obj.ccd_data
            self.unmasked_image = obj.unmasked_image
            self.unmasked_noise_map = obj.unmasked_noise_map
            self.mask_2d = obj.mask_2d
            self.image_2d = obj.image_2d
            self.noise_map_2d = obj.noise_map_2d
            self.psf = obj.psf
            self.mask_1d = obj.mask_1d
            self.image_1d = obj.image_1d
            self.noise_map_1d = obj.noise_map_1d
            self.mask_1d = obj.mask_1d
            self.sub_grid_size = obj.sub_grid_size
            self.convolver_image = obj.convolver_image
            self.uses_inversion = obj.uses_inversion
            self.convolver_mapping_matrix = obj.convolver_mapping_matrix
            self.grid_stack = obj.grid_stack
            self.padded_grid_stack = obj.padded_grid_stack
            self.border = obj.border
            self.positions = obj.positions
            self.interp_pixel_scale = obj.interp_pixel_scale
            self.uses_cluster_inversion = obj.uses_cluster_inversion
            self.cluster = obj.cluster
            self.cluster_pixel_limit = obj.cluster_pixel_limit


from autolens.data.array import grids, mask as msk
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy.plotters import galaxy_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.fixtures import *

import numpy as np

@pytest.fixture(name='galaxy_plotter_path')
def make_galaxy_plotter_setup():
    return "{}/../../../test_files/plotting/model_galaxy/".format(os.path.dirname(os.path.realpath(__file__)))


def test__intensities_is_output(gal_x1_lp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_intensities(galaxy=gal_x1_lp, grid=grid_stack_5x5.regular,
                                     mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                     positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                     output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_intensities.png' in plot_patch.paths


def test__convergence_is_output(gal_x1_mp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_convergence(galaxy=gal_x1_mp, grid=grid_stack_5x5.regular,
                                     mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                     positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                     output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_convergence.png' in plot_patch.paths


def test__potential_is_output(gal_x1_mp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_potential(galaxy=gal_x1_mp, grid=grid_stack_5x5.regular,
                                   mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                   positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                   output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_potential.png' in plot_patch.paths


def test__deflections_y_is_output(gal_x1_mp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_deflections_y(galaxy=gal_x1_mp, grid=grid_stack_5x5.regular,
                                       mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                       positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                       output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_deflections_y.png' in plot_patch.paths


def test__deflections_x_is_output(gal_x1_mp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_deflections_x(galaxy=gal_x1_mp, grid=grid_stack_5x5.regular,
                                       mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                       positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                       output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_deflections_x.png' in plot_patch.paths


def test__individual_intensities_is_output(gal_x1_lp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_intensities_subplot(galaxy=gal_x1_lp, grid=grid_stack_5x5.regular,
                                             mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                             positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                             output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_individual_intensities.png' in plot_patch.paths


def test__individual_convergence_is_output(gal_x1_lp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_convergence_subplot(galaxy=gal_x1_lp, grid=grid_stack_5x5.regular,
                                             mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                             positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                             output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_individual_convergence.png' in plot_patch.paths


def test__individual_potential_is_output(gal_x1_lp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_potential_subplot(galaxy=gal_x1_lp, grid=grid_stack_5x5.regular,
                                           mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                           positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                           output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_individual_potential.png' in plot_patch.paths


def test__individual_deflections_y_is_output(gal_x1_lp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_deflections_y_subplot(galaxy=gal_x1_lp, grid=grid_stack_5x5.regular,
                                               mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                               positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                               output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_individual_deflections_y.png' in plot_patch.paths


def test__individual_deflections_x_is_output(gal_x1_lp, grid_stack_5x5, mask_5x5, positions_5x5, galaxy_plotter_path, plot_patch):
    
    galaxy_plotters.plot_intensities_subplot(galaxy=gal_x1_lp, grid=grid_stack_5x5.regular,
                                             mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
                                             positions=positions_5x5, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                             output_path=galaxy_plotter_path, output_format='png')
    
    assert galaxy_plotter_path + 'galaxy_individual_intensities.png' in plot_patch.paths



import os
import shutil
import numpy as np
import pytest

from autolens.data.array.util import binning_util


class TestBinnedPaddingArray:

    def test__bin_up_factor_is_1__array_2d_does_not_change_shape(self):

        array_2d = np.array([[1.0, 2.0, 3.0],
                             [4.0, 5.0, 6.0],
                             [7.0, 8.0, 9.0]])

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=1)

        assert (array_2d_padded == array_2d).all()

    def test__bin_up_factor_gives_no_remainder__array_2d_does_not_change_shape(self):

        array_2d = np.ones(shape=(6, 6))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert array_2d_padded.shape == (6, 6)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert array_2d_padded.shape == (6, 6)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=6)
        assert array_2d_padded.shape == (6, 6)

        array_2d = np.ones(shape=(8, 8))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert array_2d_padded.shape == (8, 8)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=4)
        assert array_2d_padded.shape == (8, 8)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=8)
        assert array_2d_padded.shape == (8, 8)

        array_2d = np.ones(shape=(9, 9))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert array_2d_padded.shape == (9, 9)

        array_2d = np.ones(shape=(16, 16))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert array_2d_padded.shape == (16, 16)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=4)
        assert array_2d_padded.shape == (16, 16)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=8)
        assert array_2d_padded.shape == (16, 16)

        array_2d = np.ones(shape=(12, 16))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert array_2d_padded.shape == (12, 16)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=4)
        assert array_2d_padded.shape == (12, 16)

        array_2d = np.ones(shape=(16, 12))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert array_2d_padded.shape == (16, 12)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=4)
        assert array_2d_padded.shape == (16, 12)

    def test__bin_up_factor_gives_remainder__array_2d_padded_to_give_no_remainder(self):

        array_2d = np.ones(shape=(6, 6))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=4)
        assert array_2d_padded.shape == (8, 8)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=5)
        assert array_2d_padded.shape == (10, 10)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=7)
        assert array_2d_padded.shape == (7, 7)

        array_2d = np.ones(shape=(10, 10))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert array_2d_padded.shape == (12, 12)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=4)
        assert array_2d_padded.shape == (12, 12)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=6)
        assert array_2d_padded.shape == (12, 12)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=7)
        assert array_2d_padded.shape == (14, 14)

        array_2d = np.ones(shape=(7, 10))

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert array_2d_padded.shape == (9, 12)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=5)
        assert array_2d_padded.shape == (10, 10)

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=7)
        assert array_2d_padded.shape == (7, 14)

    def test__padding_using_arrays_and_not_shapes(self):

        array_2d = np.ones(shape=(4, 4))
        array_2d[1 ,1] = 2.0

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert array_2d_padded.shape == (6, 6)
        assert (array_2d_padded == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                             [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                             [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
                                             [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                             [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=5)
        assert array_2d_padded.shape == (5, 5)
        assert (array_2d_padded == np.array([[1.0, 1.0, 1.0, 1.0, 0.0],
                                             [1.0, 2.0, 1.0, 1.0, 0.0],
                                             [1.0, 1.0, 1.0, 1.0, 0.0],
                                             [1.0, 1.0, 1.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        array_2d = np.ones(shape=(2, 3))
        array_2d[1 ,1] = 2.0
        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert array_2d_padded.shape == (2, 4)
        assert (array_2d_padded == np.array([[0.0, 1.0, 1.0, 1.0],
                                             [0.0, 1.0, 2.0, 1.0]])).all()

        array_2d = np.ones(shape=(3, 2))
        array_2d[1 ,1] = 2.0
        array_2d_padded = binning_util.padded_binning_array_2d_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert array_2d_padded.shape == (4, 2)
        assert (array_2d_padded == np.array([[0.0, 0.0],
                                             [1.0, 1.0],
                                             [1.0, 2.0],
                                             [1.0, 1.0]])).all()


class TestBinnedArrays2d:

    def test__bin_using_mean__array_4x4_to_2x2__uses_mean_correctly(self):

        array_2d = np.array([[1.0, 1.0, 2.0, 2.0],
                             [1.0, 1.0, 2.0, 2.0],
                             [3.0, 3.0, 4.0, 4.0],
                             [3.0, 3.0, 4.0, 4.0]])

        binned_array_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)

        assert (binned_array_2d == np.array([[1.0, 2.0],
                                             [3.0, 4.0]])).all()

        array_2d = np.array([[1.0, 2.0, 2.0, 2.0],
                             [1.0, 6.0, 2.0, 10.0],
                             [9.0, 3.0, 4.0, 0.0],
                             [3.0, 3.0, 4.0, 4.0]])

        binned_array_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)

        assert (binned_array_2d == np.array([[2.5, 4.0],
                                             [4.5, 3.0]])).all()

    def test__bin_using_mean__array_6x3_to_2x1_and_3x6_to_1x2__uses_mean_correctly(self):

        array_2d = np.array([[1.0, 1.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)

        assert (binned_array_2d == np.array([[1.0],
                                             [2.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0],
                             [1.0, 10.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [2.0, 11.0, 2.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[2.0],
                                             [3.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[1.0, 2.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 10.0, 1.0, 11.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[2.0, 3.0]])).all()

    def test__bin_using_mean__bin_includes_padding_image_with_zeros(self):

        # Padded array:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        array_2d = np.ones(shape=(4, 4))
        array_2d[1 ,1] = 2.0
        binned_array_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[(5.0 / 9.0), (4.0 / 9.0)],
                                             [(4.0 / 9.0), (4.0 / 9.0)]])).all()

        # Padded Array:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        array_2d = np.ones(shape=(2, 3))
        array_2d[1 ,1] = 2.0
        binned_2d_array = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert (binned_2d_array == np.array([[0.5, 1.25]])).all()

    def test__bin_using_quadrature__array_4x4_to_2x2__uses_quadrature_correctly(self):

        array_2d = np.array([[1.0, 1.0, 2.0, 2.0],
                             [1.0, 1.0, 2.0, 2.0],
                             [3.0, 3.0, 4.0, 4.0],
                             [3.0, 3.0, 4.0, 4.0]])

        binned_array_2d = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)

        assert (binned_array_2d == np.array([[np.sqrt(4.0) / 4.0, np.sqrt(16.0) / 4.0],
                                             [np.sqrt(36.0) / 4.0, np.sqrt(64.0) / 4.0]])).all()

        array_2d = np.array([[1.0, 2.0, 2.0, 2.0],
                             [1.0, 6.0, 2.0, 10.0],
                             [9.0, 3.0, 4.0, 0.0],
                             [3.0, 3.0, 4.0, 4.0]])

        binned_array_2d = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)

        assert (binned_array_2d == np.array([[np.sqrt(42.0) / 4.0, np.sqrt(112.0) / 4.0],
                                             [np.sqrt(108.0) / 4.0, np.sqrt(48.0) / 4.0]])).all()

    def test__bin_using_quadrature__array_6x3_to_2x1_and_3x6_to_1x2__uses_quadrature_correctly(self):

        array_2d = np.array([[1.0, 1.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)

        assert (binned_array_2d == np.array([[np.sqrt(9.0) / 9.0],
                                             [np.sqrt(36.0) / 9.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0],
                             [1.0, 10.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [2.0, 4.0, 2.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[np.sqrt(108.0) / 9.0],
                                             [np.sqrt(48.0) / 9.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[np.sqrt(9.0) / 9.0, np.sqrt(36.0) / 9.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 10.0, 1.0, 4.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[np.sqrt(108.0) / 9.0, np.sqrt(48.0) / 9.0]])).all()

    def test__bin_using_quadrature__bin_includes_padding_image_with_zeros(self):

        # Padded array:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        array_2d = np.ones(shape=(4, 4))
        array_2d[1 ,1] = 2.0
        binned_array_2d = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[np.sqrt(7.0) / 9.0, np.sqrt(4.0) / 9.0],
                                             [np.sqrt(4.0) / 9.0, np.sqrt(4.0) / 9.0]])).all()

        # Padded Array:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        array_2d = np.ones(shape=(2, 3))
        array_2d[1 ,1] = 2.0
        binned_2d_array = binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert (binned_2d_array == np.array([[np.sqrt(2.0) / 4.0, np.sqrt(7.0) / 4.0]])).all()

    def test__bin_using_sum__array_4x4_to_2x2__uses_sum_correctly(self):

        array_2d = np.array([[1.0, 1.0, 2.0, 2.0],
                             [1.0, 1.0, 2.0, 2.0],
                             [3.0, 3.0, 4.0, 4.0],
                             [3.0, 3.0, 4.0, 4.0]])

        binned_array_2d = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)

        assert (binned_array_2d == np.array([[4.0, 8.0],
                                             [12.0, 16.0]])).all()

        array_2d = np.array([[1.0, 2.0, 2.0, 2.0],
                             [1.0, 6.0, 2.0, 10.0],
                             [9.0, 3.0, 4.0, 0.0],
                             [3.0, 3.0, 4.0, 4.0]])

        binned_array_2d = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)

        assert (binned_array_2d == np.array([[10.0, 16.0],
                                             [18.0, 12.0]])).all()

    def test__bin_using_sum__array_6x3_to_2x1_and_3x6_to_1x2__uses_sum_correctly(self):

        array_2d = np.array([[1.0, 1.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)

        assert (binned_array_2d == np.array([[9.0],
                                             [18.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0],
                             [1.0, 10.0, 1.0],
                             [1.0, 1.0, 1.0],
                             [2.0, 11.0, 2.0],
                             [2.0, 2.0, 2.0],
                             [2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[18.0],
                                             [27.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[9.0, 18.0]])).all()

        array_2d = np.array([[1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                             [1.0, 10.0, 1.0, 11.0, 2.0, 2.0],
                             [1.0, 1.0, 1.0, 2.0, 2.0, 2.0]])

        binned_array_2d = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[18.0, 27.0]])).all()

    def test__bin_using_sum__bin_includes_padding_image_with_zeros(self):

        # Padded array:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        array_2d = np.ones(shape=(4, 4))
        array_2d[1 ,1] = 2.0
        binned_array_2d = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=3)
        assert (binned_array_2d == np.array([[5.0, 4.0],
                                             [4.0, 4.0]])).all()

        # Padded Array:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        array_2d = np.ones(shape=(2, 3))
        array_2d[1 ,1] = 2.0
        binned_2d_array = binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
            array_2d=array_2d, bin_up_factor=2)
        assert (binned_2d_array == np.array([[2.0, 5.0]])).all()
        
        
class TestBinUpMask2d:

    def test__mask_4x4_to_2x2__creates_correct_binned_up_mask(self):

        mask_2d = np.array([[True, False, True, True],
                            [True, True, True, True],
                            [True, True, False, False],
                            [False, True, True, True]])

        binned_mask_2d = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_mask_2d == np.array([[False, True],
                                            [False, False]])).all()

        mask_2d = np.array([[True, True, True, True],
                            [True, True, True, True],
                            [True, True, False, False],
                            [True, True, True, True]])

        binned_mask_2d = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_mask_2d == np.array([[True, True],
                                            [True, False]])).all()

    def test__mask_6x3_to_2x1_and_3x6_to_1x2__sets_up_correct_mask(self):

        mask_2d = np.array([[True, True, True],
                            [True, True, True],
                            [True, True, True],
                            [True, True, True],
                            [True, True, True],
                            [True, True, True]])

        binned_mask_2d = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)

        assert (binned_mask_2d == np.array([[True],
                                            [True]])).all()

        mask_2d = np.array([[True, True, True],
                            [True, True, False],
                            [True, True, True],
                            [True, True, True],
                            [True, True, True],
                            [True, True, True]])

        binned_mask_2d = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
        assert (binned_mask_2d == np.array([[False],
                                            [True]])).all()

        mask_2d = np.array([[True, True, True, True, True, True],
                            [True, True, True, True, True, True],
                            [True, True, True, True, True, True]])

        binned_mask_2d = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
        assert (binned_mask_2d == np.array([[True, True]])).all()

        mask_2d = np.array([[True, True, True, True, True, True],
                            [True, True, True, True, True, True],
                            [True, True, True, True, True, False]])

        binned_mask_2d = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
        assert (binned_mask_2d == np.array([[True, False]])).all()

    def test__bin_includes_padding_image_with_zeros(self):
        # Padded mask:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        mask_2d = np.full(shape=(4, 4), fill_value=True)
        mask_2d[1, 1] = False
        mask_2d[3, 3] = False
        binned_mask_2d = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
        assert (binned_mask_2d == np.array([[False, True],
                                            [True, False]])).all()

        # Padded Array:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        mask_2d = np.full(shape=(2, 3), fill_value=True)
        mask_2d[1, 2] = False
        binned_2d_mask = binning_util.binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=2)
        assert (binned_2d_mask == np.array([[True, False]])).all()


class TestPaddedMask2dToMask1DIndex(object):

    def test__no_padding__mask_is_full_of_false__returns_indexes_in_ascending_order(self):

        mask_2d = np.full(fill_value=False, shape=(4, 4))

        padded_mask_2d_to_mask_1d_index = binning_util.padded_mask_2d_to_mask_1d_index_from_mask_2d_and_bin_up_factor(
                mask_2d=mask_2d, bin_up_factor=1)

        assert (padded_mask_2d_to_mask_1d_index == np.array([[0, 1, 2, 3],
                                                      [4, 5, 6, 7],
                                                      [8, 9, 10, 11],
                                                      [12, 13, 14, 15]])).all()

        mask_2d = np.array([[False, False],
                            [True, False],
                            [True, False]])

        padded_mask_2d_to_mask_1d_index = binning_util.padded_mask_2d_to_mask_1d_index_from_mask_2d_and_bin_up_factor(
                mask_2d=mask_2d, bin_up_factor=1)

        assert (padded_mask_2d_to_mask_1d_index == np.array([[0, 1],
                                                      [-1, 2],
                                                      [-1, 3]])).all()

    def test__includes_padding__padded_entries_are_given_minus_ones(self):

        mask_2d = np.full(fill_value=False, shape=(4, 4))

        padded_mask_2d_to_mask_1d_index = binning_util.padded_mask_2d_to_mask_1d_index_from_mask_2d_and_bin_up_factor(
                mask_2d=mask_2d, bin_up_factor=3)

        assert (padded_mask_2d_to_mask_1d_index == np.array(
            [[-1, -1, -1, -1, -1, -1],
             [-1,  0,  1,  2,  3, -1],
             [-1,  4,  5,  6,  7, -1],
             [-1,  8,  9, 10, 11, -1],
             [-1, 12, 13, 14, 15, -1],
             [-1, -1, -1, -1, -1, -1]])).all()

        mask_2d = np.array([[False, False],
                            [True, False],
                            [True, False]])

        padded_mask_2d_to_mask_1d_index = binning_util.padded_mask_2d_to_mask_1d_index_from_mask_2d_and_bin_up_factor(
                mask_2d=mask_2d, bin_up_factor=2)

        assert (padded_mask_2d_to_mask_1d_index == np.array(
            [[-1, -1],
             [0, 1],
             [-1, 2],
             [-1, 3]])).all()


class TestMask2dToBinnedMask1dIndexes:

    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask_2d = np.full(fill_value=False, shape=(4,4))

        mask_2d_to_binned_mask_1d_indexes = \
            binning_util.padded_mask_2d_to_binned_mask_1d_index_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (mask_2d_to_binned_mask_1d_indexes == np.array(
            [[0, 0, 1, 1],
            [0, 0, 1, 1],
            [2, 2, 3, 3],
            [2, 2, 3, 3]])).all()


        mask_2d = np.full(fill_value=False, shape=(9,9))

        mask_2d_to_binned_mask_1d_indexes = \
            binning_util.padded_mask_2d_to_binned_mask_1d_index_from_mask_2d_and_bin_up_factor(
                mask_2d=mask_2d, bin_up_factor=3)

        assert (mask_2d_to_binned_mask_1d_indexes ==np.array(
            [[0, 0, 0, 1, 1, 1, 2, 2, 2],
             [0, 0, 0, 1, 1, 1, 2, 2, 2],
             [0, 0, 0, 1, 1, 1, 2, 2, 2],
             [3, 3, 3, 4, 4, 4, 5, 5, 5],
             [3, 3, 3, 4, 4, 4, 5, 5, 5],
             [3, 3, 3, 4, 4, 4, 5, 5, 5],
             [6, 6, 6, 7, 7, 7, 8, 8, 8],
             [6, 6, 6, 7, 7, 7, 8, 8, 8],
             [6, 6, 6, 7, 7, 7, 8, 8, 8]])).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(self):

        mask_2d = np.array([[True, False, True, True, True, True],
                            [False, False, False, True, True, True]])

        mask_2d_to_binned_mask_1d_indexes = \
            binning_util.padded_mask_2d_to_binned_mask_1d_index_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (mask_2d_to_binned_mask_1d_indexes == np.array(
            [[-1, 0, -1, -1, -1, -1],
             [0, 0, 1, -1, -1, -1]])).all()

        mask_2d = np.array([[True, False],
                            [False, False],
                            [False, True],
                            [True, True],
                            [True, True],
                            [True, True]])

        mask_2d_to_binned_mask_1d_indexes = \
            binning_util.padded_mask_2d_to_binned_mask_1d_index_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (mask_2d_to_binned_mask_1d_indexes == np.array(
            [[-1, 0],
             [0, 0],
             [1, -1],
             [-1, -1],
             [-1, -1],
             [-1, -1]])).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask_2d = np.full(fill_value=False, shape=(5,5))

        mask_2d_to_binned_mask_1d_indexes = \
            binning_util.padded_mask_2d_to_binned_mask_1d_index_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (mask_2d_to_binned_mask_1d_indexes == np.array(
           [[-1, -1, -1, -1, -1, -1],
            [-1,  0 , 1 ,1, 2, 2],
            [-1,  3,  4, 4, 5, 5],
            [-1,  3,  4, 4, 5, 5],
            [-1,  6,  7, 7, 8, 8],
            [-1,  6,  7, 7, 8, 8]])).all()


class TestMaskedArray1DToBininedMaskedArray1d:

    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask_2d = np.full(fill_value=False, shape=(4,4))

        masked_array_1d_to_binned_masked_array_1d = \
            binning_util.masked_array_1d_to_binned_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (masked_array_1d_to_binned_masked_array_1d ==
                np.array([0, 0, 1, 1,
                          0, 0, 1, 1,
                          2, 2, 3, 3,
                          2, 2, 3, 3])).all()

        mask_2d = np.full(fill_value=False, shape=(9,9))

        masked_array_1d_to_binned_masked_array_1d = \
            binning_util.masked_array_1d_to_binned_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=3)

        assert (masked_array_1d_to_binned_masked_array_1d ==
                np.array([0, 0, 0, 1, 1, 1, 2, 2, 2,
                          0, 0, 0, 1, 1, 1, 2, 2, 2,
                          0, 0, 0, 1, 1, 1, 2, 2, 2,
                          3, 3, 3, 4, 4, 4, 5, 5, 5,
                          3, 3, 3, 4, 4, 4, 5, 5, 5,
                          3, 3, 3, 4, 4, 4, 5, 5, 5,
                          6, 6, 6, 7, 7, 7, 8, 8, 8,
                          6, 6, 6, 7, 7, 7, 8, 8, 8,
                          6, 6, 6, 7, 7, 7, 8, 8, 8])).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(self):

        mask_2d = np.array([[True, False, True, True, True, True],
                            [False, False, False, True, True, True]])

        masked_array_1d_to_binned_masked_array_1d = \
            binning_util.masked_array_1d_to_binned_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (masked_array_1d_to_binned_masked_array_1d == np.array(
            [0, 0, 0, 1])).all()

        mask_2d = np.array([[True, False],
                            [False, False],
                            [False, True],
                            [True, True],
                            [True, False],
                            [True, False]])

        masked_array_1d_to_binned_masked_array_1d = \
            binning_util.masked_array_1d_to_binned_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (masked_array_1d_to_binned_masked_array_1d == np.array(
            [0, 0, 0, 1, 2, 2])).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask_2d = np.full(fill_value=False, shape=(5,5))

        masked_array_1d_to_binned_masked_array_1d = \
            binning_util.masked_array_1d_to_binned_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (masked_array_1d_to_binned_masked_array_1d == np.array(
            [0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 6, 7, 7, 8, 8])).all()


class TestBinnedMaskArrayToMaskedArray:

    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask_2d = np.full(fill_value=False, shape=(4,4))

        binned_masked_array_1d_to_masked_array_1d = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_to_masked_array_1d == np.array(
            [0, 2, 8, 10])).all()

        mask_2d = np.full(fill_value=False, shape=(9,9))

        binned_masked_array_1d_to_masked_array_1d = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=3)

        assert (binned_masked_array_1d_to_masked_array_1d == np.array(
            [0, 3, 6, 27, 30, 33, 54, 57, 60])).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(self):

        mask_2d = np.array([[True, False, True, True, True, True],
                            [False, False, False, True, True, True]])

        binned_masked_array_1d_to_masked_array_1d = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_to_masked_array_1d == np.array([0, 3])).all()

        mask_2d = np.array([[True, False],
                            [False, False],
                            [False, True],
                            [True, True],
                            [True, True],
                            [True, True]])

        binned_masked_array_1d_to_masked_array_1d = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_to_masked_array_1d == np.array([0, 3])).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask_2d = np.full(fill_value=False, shape=(5,5))

        binned_masked_array_1d_to_masked_array_1d = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_to_masked_array_1d == np.array([0, 1, 3, 5, 6, 8, 15, 16, 18])).all()
        
        
class TestBinnedMaskArrayToMaskedArrayAll:

    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask_2d = np.full(fill_value=False, shape=(4,4))

        binned_masked_array_1d_to_masked_array_1d_all, binned_masked_array_1d_sizes = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_all_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_sizes == np.array([4, 4, 4, 4])).all()

        assert (binned_masked_array_1d_to_masked_array_1d_all == np.array(
            [[0,   1,  4,  5],
             [2,   3,  6,  7],
             [8,   9, 12, 13],
             [10, 11, 14, 15]])).all()

        mask_2d = np.full(fill_value=False, shape=(9,9))

        binned_masked_array_1d_to_masked_array_1d_all, binned_masked_array_1d_sizes = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_all_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=3)

        assert (binned_masked_array_1d_sizes == np.array([9, 9, 9, 9, 9, 9, 9, 9, 9])).all()

        assert (binned_masked_array_1d_to_masked_array_1d_all == np.array(
            [[ 0,  1,  2,  9, 10, 11, 18, 19, 20],
             [ 3,  4,  5, 12, 13, 14, 21, 22, 23],
             [ 6,  7,  8, 15, 16, 17, 24, 25, 26],
             [27, 28, 29, 36, 37, 38, 45, 46, 47],
             [30, 31, 32, 39, 40, 41, 48, 49, 50],
             [33, 34, 35, 42, 43, 44, 51, 52, 53],
             [54, 55, 56, 63, 64, 65, 72, 73, 74],
             [57, 58, 59, 66, 67, 68, 75, 76, 77],
             [60, 61, 62, 69, 70, 71, 78, 79, 80]])).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(self):

        mask_2d = np.array([[True, False, True, True, True, True],
                            [False, False, False, True, True, True]])

        binned_masked_array_1d_to_masked_array_1d_all, binned_masked_array_1d_sizes = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_all_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_sizes == np.array([3, 1])).all()

        assert (binned_masked_array_1d_to_masked_array_1d_all == np.array(
            [[0, 1, 2, -1],
             [3, -1, -1, -1]])).all()

        mask_2d = np.array([[True, False],
                            [False, False],
                            [False, True],
                            [True, True],
                            [True, True],
                            [True, True]])

        binned_masked_array_1d_to_masked_array_1d_all, binned_masked_array_1d_sizes = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_all_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_sizes == np.array([3, 1])).all()

        assert (binned_masked_array_1d_to_masked_array_1d_all == np.array(
            [[0, 1, 2, -1],
            [3, -1, -1, -1]])).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask_2d = np.full(fill_value=False, shape=(5,5))

        binned_masked_array_1d_to_masked_array_1d_all, binned_masked_array_1d_sizes = \
            binning_util.binned_masked_array_1d_to_masked_array_1d_all_from_mask_2d_and_bin_up_factor(
            mask_2d=mask_2d, bin_up_factor=2)

        assert (binned_masked_array_1d_sizes == np.array([1, 2, 2, 2, 4, 4, 2, 4, 4])).all()

        assert (binned_masked_array_1d_to_masked_array_1d_all == np.array(
            [[0, -1, -1, -1],
             [1, 2, -1, -1],
             [3, 4, -1, -1],
             [5, 10, -1, -1],
             [6, 7, 11, 12],
             [8, 9, 13, 14],
             [15, 20, -1, -1],
             [16, 17, 21, 22],
             [18, 19, 23, 24]])).all()


from autolens.lens.plotters import plane_plotters
from test.fixtures import *


@pytest.fixture(name='plane_plotter_path')
def make_plane_plotter_setup():
    return "{}/../../test_files/plotting/plane/".format(os.path.dirname(os.path.realpath(__file__)))


def test__all_individual_plotters__output_file_with_default_name(
        plane_5x5, mask_5x5, positions_5x5, plane_plotter_path, plot_patch):

    plane_plotters.plot_image_plane_image(
        plane=plane_5x5, mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
        positions=positions_5x5,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=plane_plotter_path, output_format='png')

    assert plane_plotter_path + 'plane_image_plane_image.png' in plot_patch.paths

    plane_plotters.plot_plane_image(
        plane=plane_5x5, positions=positions_5x5,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=plane_plotter_path, output_format='png')

    assert plane_plotter_path + 'plane_image.png' in plot_patch.paths

    plane_plotters.plot_convergence(
        plane=plane_5x5, mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=plane_plotter_path, output_format='png')

    assert plane_plotter_path + 'plane_convergence.png' in plot_patch.paths

    plane_plotters.plot_potential(
        plane=plane_5x5, mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=plane_plotter_path, output_format='png')

    assert plane_plotter_path + 'plane_potential.png' in plot_patch.paths

    plane_plotters.plot_deflections_y(
        plane=plane_5x5, mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=plane_plotter_path, output_format='png')

    assert plane_plotter_path + 'plane_deflections_y.png' in plot_patch.paths

    plane_plotters.plot_deflections_x(
        plane=plane_5x5, mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=plane_plotter_path, output_format='png')

    assert plane_plotter_path + 'plane_deflections_x.png' in plot_patch.paths

    plane_plotters.plot_plane_grid(
        plane=plane_5x5, output_path=plane_plotter_path, output_format='png')

    assert plane_plotter_path + 'plane_grid.png' in plot_patch.paths



from autolens.lens.plotters import lens_fit_plotters
from test.fixtures import *


@pytest.fixture(name='lens_fit_plotter_path')
def make_lens_fit_plotter_setup():
    return "{}/../../test_files/plotting/fit/".format(os.path.dirname(os.path.realpath(__file__)))


def test__fit_sub_plot(lens_fit_x2_plane_5x5, lens_fit_plotter_path, plot_patch):

    lens_fit_plotters.plot_fit_subplot(
        fit=lens_fit_x2_plane_5x5,
        should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'lens_fit.png' in plot_patch.paths

def test__fit_for_plane_subplot(
        lens_fit_x1_plane_5x5, lens_fit_x2_plane_5x5, lens_fit_plotter_path, plot_patch):

    lens_fit_plotters.plot_fit_subplot_for_plane(
        fit=lens_fit_x1_plane_5x5, plane_index=0,
        should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'lens_fit_plane_0.png' in plot_patch.paths

    lens_fit_plotters.plot_fit_subplot_for_plane(
        fit=lens_fit_x2_plane_5x5, plane_index=0,
        should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'lens_fit_plane_0.png' in plot_patch.paths

    lens_fit_plotters.plot_fit_subplot_for_plane(
        fit=lens_fit_x2_plane_5x5, plane_index=1,
        should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'lens_fit_plane_1.png' in plot_patch.paths

def test__fit_for_planes_subplot(
        lens_fit_x1_plane_5x5, lens_fit_x2_plane_5x5, lens_fit_plotter_path, plot_patch):

    lens_fit_plotters.plot_fit_subplot_of_planes(
        fit=lens_fit_x1_plane_5x5,
        should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'lens_fit_plane_0.png' in plot_patch.paths

    lens_fit_plotters.plot_fit_subplot_of_planes(
        fit=lens_fit_x2_plane_5x5,
        should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'],
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'lens_fit_plane_0.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'lens_fit_plane_1.png' in plot_patch.paths

def test__fit_individuals__source_and_lens__depedent_on_input(
        lens_fit_x1_plane_5x5, lens_fit_x2_plane_5x5, lens_fit_plotter_path, plot_patch):

    lens_fit_plotters.plot_fit_individuals(
        fit=lens_fit_x1_plane_5x5,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_chi_squared_map=True,
        should_plot_subtracted_images_of_planes=True,
        should_plot_model_images_of_planes=True,
        should_plot_plane_images_of_planes=True,
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'fit_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_signal_to_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_model_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_residual_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_normalized_residual_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_subtracted_image_of_plane_0.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_model_image_of_plane_0.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_plane_image_of_plane_0.png' in plot_patch.paths

    lens_fit_plotters.plot_fit_individuals(
        fit=lens_fit_x2_plane_5x5,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_chi_squared_map=True,
        should_plot_subtracted_images_of_planes=True,
        should_plot_model_images_of_planes=True,
        should_plot_plane_images_of_planes=True,
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'fit_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_signal_to_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_model_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_residual_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_normalized_residual_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_subtracted_image_of_plane_0.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_subtracted_image_of_plane_1.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_model_image_of_plane_0.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_model_image_of_plane_1.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_plane_image_of_plane_0.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_plane_image_of_plane_1.png' in plot_patch.paths


from autolens import text_util


def test__within_radius_label_value_and_unit_string():
    string0 = text_util.within_radius_label_value_and_unit_string(
        prefix='mass', radius=1.0, unit_length='arcsec', value=30.0,
        unit_value='solMass', whitespace=40)

    string1 = text_util.within_radius_label_value_and_unit_string(
        prefix='mass', radius=1.0, unit_length='arcsec', value=30.0,
        unit_value='solMass', whitespace=35)

    string2 = text_util.within_radius_label_value_and_unit_string(
        prefix='mass', radius=1.0, unit_length='arcsec', value=30.0,
        unit_value='solMass', whitespace=30)

    assert string0 == 'mass_within_1.00_arcsec                 3.0000e+01 solMass'
    assert string1 == 'mass_within_1.00_arcsec            3.0000e+01 solMass'
    assert string2 == 'mass_within_1.00_arcsec       3.0000e+01 solMass'

    string = text_util.within_radius_label_value_and_unit_string(
        prefix='mass', radius=1.0, unit_length='arcsec2', value=40.0,
        unit_value='solMass2', whitespace=40)

    assert string == 'mass_within_1.00_arcsec2                4.0000e+01 solMass2'




class MockLensData(object):

    def __init__(self, ccd_data, mask, grid_stack, padded_grid_stack, border, convolver_image,
                 convolver_mapping_matrix, cluster):

        self.ccd_data = ccd_data

        self.unmasked_image = ccd_data.image
        self.unmasked_noise_map = ccd_data.noise_map
        self.pixel_scale = ccd_data.pixel_scale

        self.psf = ccd_data.psf

        self.mask_2d = mask
        self.mask_1d = self.mask_2d.map_2d_array_to_masked_1d_array(array_2d=self.mask_2d)

        self.grid_stack = grid_stack
        self.padded_grid_stack = padded_grid_stack
        self.sub_grid_size = self.grid_stack.sub.sub_grid_size
        self.border = border
        self.convolver_image = convolver_image
        self.convolver_mapping_matrix = convolver_mapping_matrix

        self.image_1d = self.mask_2d.map_2d_array_to_masked_1d_array(array_2d=self.unmasked_image)
        self.noise_map_1d = self.mask_2d.map_2d_array_to_masked_1d_array(array_2d=self.unmasked_noise_map)

        self.image_2d = self.map_to_scaled_array(array_1d=self.image_1d)
        self.noise_map_2d = self.map_to_scaled_array(array_1d=self.noise_map_1d)
        self.positions = None

        self.cluster = cluster

        self.uses_cluster_inversion = False

    @property
    def array_1d_from_array_2d(self):
        return self.grid_stack.regular.array_1d_from_array_2d

    @property
    def map_to_scaled_array(self):
        return self.grid_stack.scaled_array_2d_from_array_1d



import autofit as af


def within_radius_label_value_and_unit_string(prefix, radius, unit_length, value,
                                              unit_value, whitespace):
    label = prefix + '_within_{:.2f}_{}'.format(radius, unit_length)
    return af.text_util.label_value_and_unit_string(
        label=label, value=value, unit=unit_value, whitespace=whitespace)
